import datetime
from abc import ABC, abstractmethod
from typing import Dict, List, Union
from datalogue.errors import DtlError
from datalogue.dtl_utils import _parse_string_list


class ShardIterator(ABC):
    """
    Abstract class containing the five subclasses:
      TrimHorizon, for beginning at the beginning of the stream
      Latest, for beginning at the most recent record
      AtTimestamp, for beginning at a specified time
      AtSequenceNumber, for beginning at a specified sequence number
      AfterSequenceNumber, for beginning just after a specified sequence number
    """

    type_field = "type"

    type_str = ""

    def __init__(self, type: str):
        self.type = type
        super().__init__()

    def _base_payload(self) -> dict:
        base = [(ShardIterator.type_field, self.type_str)]
        return dict(base)

    @abstractmethod
    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        pass

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, "ShardIterator"]:
        type_field = d.get(ShardIterator.type_field)
        if not isinstance(type_field, str):
            return DtlError("The type is missing from JSON")
        elif type_field == TrimHorizon.type_str:
            return TrimHorizon()
        elif type_field == Latest.type_str:
            return Latest()
        elif type_field == AtSequenceNumber.type_str:
            return AtSequenceNumber._from_payload(d)
        elif type_field == AfterSequenceNumber.type_str:
            return AfterSequenceNumber._from_payload(d)
        elif type_field == AtTimestamp.type_str:
            return AtTimestamp._from_payload(d)
        else:
            return DtlError("The object %s is not a ShardIterator definition" % str(d))


class TrimHorizon(ShardIterator):
    """
      Begins reading stream from the beginning
    """

    type_str = "TrimHorizon"

    def __init__(self):
        super().__init__(TrimHorizon.type_str)

    def __repr__(self):
        return f"{self.__class__.__name__}"

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        return base


class Latest(ShardIterator):
    """
      Begins reading stream from the beginning
    """

    type_str = "Latest"

    def __init__(self):
        super().__init__(Latest.type_str)

    def __repr__(self):
        return f"{self.__class__.__name__}"

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        return base


class AtSequenceNumber(ShardIterator):
    """
      Begins reading stream starting from element of specified sequence number
    """

    type_str = "AtSequenceNumber"

    def __init__(self, sequence_number: str):
        self.sequence_number = sequence_number
        super().__init__(AtSequenceNumber.type_str)

    def __repr__(self):
        return f"{self.__class__.__name__}(sequence_number= {self.sequence_number})"

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        base["sequenceNumber"] = self.sequence_number
        return base

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, "AtSequenceNumber"]:
        sequence_number = d.get("sequenceNumber")
        if sequence_number is None:
            return DtlError("The `sequenceNumber` is missing from JSON")
        return AtSequenceNumber(sequence_number)


class AfterSequenceNumber(ShardIterator):
    """
      Begins reading stream starting immediately after element of specified sequence number
    """

    type_str = "AfterSequenceNumber"

    def __init__(self, sequence_number: str):
        self.sequence_number = sequence_number
        super().__init__(AfterSequenceNumber.type_str)

    def __repr__(self):
        return f"{self.__class__.__name__}(sequence_number= {self.sequence_number})"

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        base["sequenceNumber"] = self.sequence_number
        return base

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, "AfterSequenceNumber"]:
        sequence_number = d.get("sequenceNumber")
        if sequence_number is None:
            return DtlError("The `sequenceNumber` is missing from JSON")
        return AfterSequenceNumber(sequence_number)


class AtTimestamp(ShardIterator):
    """
      Begins reading stream from specified timestamp
    """

    type_str = "AtTimestamp"

    def __init__(self, timestamp: int):
        self.timestamp = timestamp
        super().__init__(AtTimestamp.type_str)

    def __repr__(self):
        return f"{self.__class__.__name__}(timestamp= {self.timestamp})"

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        base["timestamp"] = self.timestamp
        return base

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, "AtTimestamp"]:
        timestamp = d.get("timestamp")
        if timestamp is None:
            return DtlError("The `timestamp` is missing from JSON")
        timestamp = datetime.datetime.fromtimestamp(timestamp, tz=datetime.timezone.utc)
        if not isinstance(timestamp, datetime.datetime):
            return DtlError("The `timestamp` parsed from JSON is not a datetime format.")

        return AtTimestamp(timestamp)


class ShardAttributes(object):
    """
    Configures the shard attributes for the streaming source
    :param partition_keys can be used to specify to read from a specific partition key or partition keys. If unspecified, reads from all partition keys (consequently, all shards) of the stream
    :param shard_iterator: iterator type used for this source, as a subclass of ShardIterator
    :param refresh_interval: time between requests for data loads, in milliseconds
    :param limit: maximum number of transactions ingested from stream per second (for Kinesis, max supported is 5 and each transaction can contain 10k records)
    """

    def __init__(
        self, shard_iterators: Dict[str, ShardIterator] = {}, refresh_interval: float = 1, limit: int = 5,
    ):
        self.shard_iterators = shard_iterators
        self.refresh_interval = refresh_interval
        self.limit = limit

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(shard_iterators= {self.shard_iterators}, "
            f"refresh_interval= {self.refresh_interval}, limit= {self.limit})"
        )

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        json = {}
        json["shardIterators"] = {
            shard_id: ShardIterator._as_payload(shard_iter) for shard_id, shard_iter in self.shard_iterators.items()
        }
        json["refreshInterval"] = self.refresh_interval
        json["limit"] = self.limit
        return json

    @staticmethod
    def _from_payload(shard_attributes: dict) -> Union[DtlError, "ShardAttributes"]:
        shard_iterators = {
            shard_id: ShardIterator._from_payload(shard_iter)
            for shard_id, shard_iter in shard_attributes.get("shardIterators", {}).items()
        }
        refresh_interval = shard_attributes["refreshInterval"]
        limit = shard_attributes["limit"]
        return ShardAttributes(shard_iterators, refresh_interval, limit)



from datetime import datetime
from typing import Optional, Union, List
from uuid import UUID

from datalogue.errors import DtlError
from datalogue.dtl_utils import map_option, _parse_uuid, _parse_datetime


class Tag:
    """
    A unique tag of metadata that can be applied to datastores for targeted collection, viewing, and action orchestrations
    """

    def __init__(
        self,
        name: str,
        id: Optional[UUID] = None,
        created_at: Optional[datetime] = None,
        created_by: Optional[UUID] = None,
    ):
        self.id = id
        self.name = name
        self.created_at = created_at
        self.created_by = created_by

    def __eq__(self, other):
        return (
            (self.name == other.name)
            and (self.created_at == other.created_at)
            and (self.created_by == other.created_by)
        )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(\n"
            f"id= {self.id!r},\n "
            f"name= {self.name!r},\n"
            f"created_at= {self.created_at!r},\n"
            f"created_by= UUID({self.created_by!r})\n"
            f")"
        )

    @staticmethod
    def _from_payload(payload: dict) -> Union[DtlError, "Tag"]:
        name = payload.get("name")
        if name is None:
            return DtlError("'name' is missing for a tag")

        id = map_option(payload.get("id"), _parse_uuid)
        created_at = payload.get("createdAt")
        created_by = payload.get("createdBy")
        if isinstance(created_at, DtlError):
            return created_at

        return Tag(name, id, created_at, created_by)



from typing import Union, List, Callable
from uuid import UUID

from datalogue.clients._http import _HttpClient, HttpMethod, MultiPartData, MultiPartFile
from datalogue.dtl_utils import _parse_list, _parse_uuid
from datalogue.errors import DtlError
from datalogue.models.ontology import Ontology, OntologyNode, ClassMapping
from datalogue.models.permission import OntologyPermission
from datalogue.models.scope_level import Scope


class OntologyClient:
    """
    Client to interact with the ontology
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/yggy"
        self.mappings_uri = "/scout/classes/mappings"

    def create(self, ontology: Ontology) -> Union[DtlError, Ontology]:
        """
        Create :class:`Ontology` object given ontology object
        :return: :class:`Ontology` object, else returns :class:`DtlError`
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/ontology/import", HttpMethod.POST, body=Ontology._as_payload(ontology),
        )

        if isinstance(res, DtlError):
            return res

        return Ontology._from_payload(res)

    def get(self, ontology_id: UUID) -> Union[DtlError, Ontology]:
        """
        Get :class:`Ontology` object given ontology_id
        :return: :class:`Ontology` object, else returns :class:`DtlError`
        """
        res = self.http_client.make_authed_request(self.service_uri + f"/ontology/{str(ontology_id)}", HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return Ontology._from_payload(res)

    def update(self, ontology_id: UUID, updated_ontology: Ontology) -> Union[DtlError, Ontology]:
        """
        Update :class:`Ontology` method that will replace the existing ontology.
        The structure of this ontology will replace the existing structure. Please make sure you include a full
        ontological structure and not just the changes/modifications
        :param ontology_id: UUID is the id of the ontology that you want to replace/modify/update.
        :param updated_ontology: Ontology object that will replace the existing ontology.
        :return: The updated Ontology if successful, else returns :class:`DtlError`
        """
        res = self.http_client.make_authed_request(
            self.service_uri + f"/ontology/{str(ontology_id)}",
            HttpMethod.PUT,
            body=Ontology._as_payload(updated_ontology),
        )

        if isinstance(res, DtlError):
            return res

        return Ontology._from_payload(res)

    def delete(self, ontology_id: UUID) -> Union[DtlError, bool]:
        """
        Delete :class:`Ontology` based on the given ontology_id
        :return: True if successful, else returns :class:`DtlError`
        """
        res = self.http_client.make_authed_request(
            self.service_uri + f"/ontology/{str(ontology_id)}", HttpMethod.DELETE
        )
        if isinstance(res, DtlError):
            return res
        else:
            return True

    def get_class(self, class_id: UUID) -> Union[DtlError, OntologyNode]:
        """
        Retrieve an ontological class locally/ a class is represented as an OntologyNode object

        :param class_id: id of the ontological class to retrieve
        :return: the OntologyNode object corresponding to the id
        """
        res = self.http_client.make_authed_request(f"/yggy/entity/{class_id}", HttpMethod.GET)
        if isinstance(res, DtlError):
            return res

        return OntologyNode.from_payload(res)

    def share(
        self, ontology_id: UUID, target_id: UUID, target_type: Scope, permission: OntologyPermission,
    ) -> Union[DtlError, bool]:
        """
        Share the given ontology with an specific user or group with the desired permission (Write or Read)

        :param ontology_id: UUID is the id of the ontology that you want to share
        :param target_id: UUID is the id of the User the Group or Organization you want to share with (depending on the target_type param)
        :param target_type: Scope (`Group`, `User` or `Organization`) with whom you want to share the ontology. It can be a User, Group or Organization.
        :param permission: OntologyPermission (`Write` or `Read`) the permission you want to grant
        :return: True if successful, else returns :class:`DtlError`
        """

        body = Ontology._create_body_for_sharing(target_id, target_type, permission)
        if isinstance(body, DtlError):
            return body

        res = self.http_client.make_authed_request(
            self.service_uri + f"/ontology/{str(ontology_id)}/share", HttpMethod.POST, body,
        )
        if isinstance(res, DtlError):
            return res
        else:
            return True

    def unshare(self, ontology_id: UUID, target_id: UUID, target_type: Scope) -> Union[DtlError, bool]:
        """
        Unshare the given ontology from an specific user or group. That User/Group won't be able to access that ontology
        any more.

        :param ontology_id: UUID is the id of the ontology that you want to unshare
        :param target_id: UUID is the id of the User, the Group or Organization you want to unshare from (depending on the target_type param)
        :param target_type: Scope (`Group`, `User` or `Organization`) with whom you want to unshare the ontology. It can be a User, Group or Organization.
        :return: True if successful, else returns :class:`DtlError`
        """
        body = Ontology._create_body_for_unsharing(target_id, target_type, "None")

        if isinstance(body, DtlError):
            return body

        res = self.http_client.make_authed_request(
            self.service_uri + f"/ontology/{str(ontology_id)}/share", HttpMethod.POST, body,
        )
        if isinstance(res, DtlError):
            return res
        else:
            return True

    def list(self, page: int = 1, item_per_page: int = 25) -> Union[DtlError, List[Ontology]]:
        """
        List the ontologies

        :param page: page to be retrieved
        :param item_per_page: number of ontologies to be put in a page
        :return: Returns a List of all the available Ontologies or an error message as a string

        """
        res = self.http_client.make_authed_request(
            self.service_uri + f"/ontologies?page={page}&size={item_per_page}", HttpMethod.GET,
        )

        if isinstance(res, DtlError):
            return res

        return _parse_list(Ontology._from_payload)(res)

    def overwrite(self, ontology_id: UUID, ontology: Ontology) -> Union[DtlError, Ontology]:
        """
        It overwrites all ontology and node information for the given ontology id including training data if there is an ontology for the given id. 
        Otherwise, it creates brand new ontology with the given ontology id
        :return: :class:`Ontology` object, else returns :class:`DtlError`
        """
        res = self.http_client.make_authed_request(
            self.service_uri + f"/ontology/{ontology_id}/import", HttpMethod.PUT, body=Ontology._as_payload(ontology),
        )

        if isinstance(res, DtlError):
            return res

        return Ontology._from_payload(res)

    def get_class_mappings(self, class_ids: List[UUID]) -> Union[DtlError, List[ClassMapping]]:
        """
        Retrieve source and destination mappings of given classes
        :param class_ids:  UUID of classes
        :return: List of ClassMappings
        """
        req_body = []
        for i in class_ids:
            req_body.append(str(i))

        res = self.http_client.make_authed_request(self.mappings_uri, HttpMethod.POST, body=req_body)
        if isinstance(res, DtlError):
            return res

        return _parse_list(ClassMapping.from_payload)(res)

    ####### Internal functions
    def __iterate_over_tree(
        self, current_node: OntologyNode, f: Callable[[OntologyNode], OntologyNode]
    ) -> OntologyNode:
        if not current_node.children:
            updated_node = f(current_node)
            return updated_node
        else:
            updated_children = list(map(lambda c: self.__iterate_over_tree(c, f), current_node.children))
            updated_node = f(current_node)
            updated_node.children = updated_children
            return updated_node

    def __get_node_dataset_ids(self, node_id: UUID) -> List[str]:
        entity_res = self.http_client.make_authed_request(f"/yggy/entity/{node_id}", HttpMethod.GET)
        training_data_list = entity_res.get("trainingDataInfo")
        if training_data_list is None:
            training_data_list = []

        dataset_ids = []
        for trainingData in training_data_list:
            dataset_ids.append(trainingData["datasetId"])

        return dataset_ids

    def __get_node_path(self, node_id: UUID, dataset_id: str) -> List[str]:
        entity_res = self.http_client.make_authed_request(f"/yggy/entity/{node_id}", HttpMethod.GET)
        training_data_list = entity_res.get("trainingDataInfo")
        if training_data_list is None:
            training_data_list = []

        path = []
        for trainingData in training_data_list:
            if dataset_id == trainingData.get("datasetId"):
                path = trainingData.get("nodePath")

        return path



from abc import ABC, abstractmethod
from typing import List, Union, Optional
from datalogue.models.datastore import (
    Datastore,
    DatastoreDef,
    _datastore_def_from_payload,
    CredentialType,
)
from datalogue.models.transformations import (
    Transformation,
    _transformation_from_payload,
)
from datalogue.models.transformations.commons import DataType
from datalogue.models.transformations.structure import Structure, PathNodeDescription
from datalogue.models.pipeline_behavior import (
    PipelineBehavior,
    Write,
    _pipeline_behavior_from_payload,
)
from datalogue.dtl_utils import _parse_list, _parse_uuid, map_option
from datalogue.errors import DtlError, _property_not_found
from uuid import UUID


class Definition:
    """
    A Pipeline is a series of transformations that will output its results to a target
    and can contain Child(ren ?) Pipelines. cf diagram below.

        Parent Pipelines -> Transformations -+-> Target
                                             |
                                             +-> Children Pipelines
    """

    def __init__(
        self,
        transformations: List[Transformation],
        forks: List["Definition"],
        target: Union[Datastore, DatastoreDef],
        pipeline_behavior: PipelineBehavior = Write(),
    ):
        if isinstance(target, Datastore):
            target = target.definition

        self.transformations = transformations
        self.forks = forks
        self.target = target
        self.pipeline_behavior = pipeline_behavior

    def __repr__(self):
        return (
            f"Definition(transformations= {self.transformations!r}, forks= {self.forks!r}, target= {self.target!r}"
            f", pipeline_behavior= {self.pipeline_behavior})"
        )

    def __eq__(self, other: "Definition"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def _as_payload(self) -> Union[DtlError, dict]:
        if self.target.datastore_id is None:
            return DtlError("Cannot serialize a pipeline with a target that was not saved to the database (id missing)")

        if self.target.type != CredentialType.JDBC and not isinstance(self.pipeline_behavior, Write):
            return DtlError("Only JDBC targets can have pipeline behavior other than 'Write'")

        target_payload = self.target._as_payload()
        target_payload["pipelineBehavior"] = self.pipeline_behavior._as_payload()
        return {
            "transformations": list(map(lambda s: s._as_payload(), self.transformations)),
            "pipelines": list(map(lambda s: s._as_payload(), self.forks)),
            "target": target_payload,
        }

    @staticmethod
    def _from_payload(payload: dict) -> Union[DtlError, "Definition"]:
        """
        Builds a pipeline object from a dictionary,

        :param payload: dictionary parsed from json
        :return:
        """

        transformations = payload.get("transformations")
        if transformations is not None:
            transformations = _parse_list(_transformation_from_payload)(transformations)
            if isinstance(transformations, DtlError):
                return transformations
        else:
            transformations = list()

        forks = payload.get("pipelines")
        if forks is not None:
            forks = _parse_list(Definition._from_payload)(forks)
            if isinstance(forks, DtlError):
                return forks
        else:
            forks = list()

        target = payload.get("target")
        if target is None:
            return DtlError("Cannot have a pipeline without a 'target' property")
        else:
            target = _datastore_def_from_payload(target)
            if isinstance(target, DtlError):
                return target

        pipeline_behavior = payload.get("target").get("pipelineBehavior")
        if pipeline_behavior is not None:
            pipeline_behavior = _pipeline_behavior_from_payload(pipeline_behavior)
            if isinstance(pipeline_behavior, DtlError):
                return pipeline_behavior
        else:
            pipeline_behavior = Write()

        return Definition(transformations, forks, target, pipeline_behavior)


class EnvVariable:
    """
    Object that represents an environment variable.

    An environment variable can be either a literal or an expression that returns a value.

    Right now you can input as a value:
        - a string: "a"
        - an int: 3
        - a float: 4.3
        - a boolean: false
        - a script to index a source of data: `Source(<source_json_object>).index(n => n.label == "some_label")`. This
           will cache the data into a `Map` with the key being the value of the label you specified and value being
           the whole adg.

    The environment variables are injected as arguments into the lambda functions defined inside transformations.

        EX: MapFunction transformation

    ALPHA FEATURE
    """

    def __init__(self, data_type: DataType, value: str, key: str):
        """
        Builds an environment variable to be evaluated before execution of a stream

        :param data_type: type of the variable
        :param value: string to be evaluated
        :param key: key to be used to retrieve the value inside a lambda function
        """
        self.type = data_type
        self.value = value
        self.key = key

    def __eq__(self, other: "EnvVariable"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"EnvVariable(type= {self.type.value}, value= {self.value!r}, key= {self.key!r})"

    def _as_payload(self):
        return {"type": self.type.value, "value": self.value, "key": self.key}

    @staticmethod
    def _from_payload(payload: dict) -> Union["EnvVariable", DtlError]:

        data_type = payload.get("type")
        if data_type is None:
            return _property_not_found("type", payload)

        data_type = DataType.from_str(data_type)
        if isinstance(data_type, DtlError):
            return data_type

        value = payload.get("value")
        if value is None:
            return _property_not_found("value", payload)

        key = payload.get("key")
        if key is None:
            return _property_not_found("key", payload)

        return EnvVariable(data_type, value, key)


class PipelineDef:
    """
    Describes a stream of data flowing from a source to one or several destinations with transformations

        Source -+-> Pipeline 0
                |
                +-> Pipeline 1

               ...
    """

    def __init__(
        self,
        source: Union[DatastoreDef, Datastore],
        definitions: List[Definition],
        env: Optional[List[EnvVariable]] = None,
    ):
        """
        Creates a forks from a source flowing into different definitions

        :param source: source of the streaming data
        :param definitions: pipelines containing the transformations and targets for the data
        :param env: list of environment variables to be evaluated before beginning streaming the data
        """
        if isinstance(source, Datastore):
            source = source.definition

        self.source = source
        self.definitions = definitions
        self.env = env

    def __repr__(self):
        return f"PipelineDef(source= {self.source!r}, definitions= {self.definitions!r}, env= {self.env!r})"

    def __eq__(self, other: "PipelineDef"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def _as_payload(self):
        base = {
            "source": self.source._as_payload(),
            "pipelines": list(map(lambda s: s._as_payload(), self.definitions)),
        }
        if self.env is not None:
            base["env"] = list(map(lambda s: s._as_payload(), self.env))

        return base

    @staticmethod
    def _from_payload(payload: dict) -> Union[DtlError, "PipelineDef"]:
        """
        Builds a PipelineDef instance from a payload object

        :param payload:
        :return: if fails returns a string with the error message
        """
        source = payload.get("source")
        if source is None:
            return DtlError("stream needs a source of data")
        else:
            source = _datastore_def_from_payload(source)
            if isinstance(source, DtlError):
                return source

        pipelines = payload.get("pipelines")
        if pipelines is None:
            return DtlError("streams needs a 'pipelines' property")
        else:
            pipelines = _parse_list(Definition._from_payload)(pipelines)
            if isinstance(pipelines, DtlError):
                return pipelines

        env = payload.get("env")
        if env is not None:
            env = _parse_list(EnvVariable._from_payload)(env)
            if isinstance(env, DtlError):
                return env

        return PipelineDef(source, pipelines, env)

    @staticmethod
    def simple_builder(
        source_datastore: Datastore,
        transformation_list: List[Transformation],
        destination_datastore: Datastore,
        prepend_structure: bool = False,
    ) -> Union["PipelineDef", DtlError]:
        """
        Helper to create a simple PipelineDef from one source to one destination, optionally with 
        transformations
        :param source_datastore: the existing datastore to be used as pipeline source
        :param transformation_list: a list of transformations 
        :param destination_datastore: the existing datastore to be used as pipeline destination
        :param prepend_structure: constructs a leading Structure transformation based on source 
        fields
        """
        if prepend_structure == True:
            field_list = []
            if source_datastore.schema_paths:
                for path in source_datastore.schema_paths:
                    field_list.append(PathNodeDescription([path[-1]], [path[-1]]))
                transformations = [Structure(field_list)] + transformation_list
            else:
                return DtlError(
                    "Could not find fields within source "
                    + source_datastore.name
                    + " to build the leading Structure transformation"
                )
        else:
            transformations = transformation_list
        return PipelineDef(
            source=source_datastore,
            definitions=[Definition(transformations=transformations, forks=[], target=destination_datastore)],
        )


class Pipeline:
    def __init__(
        self, id: UUID, name: str, pipeline_def: PipelineDef, warnings: List[str] = [], is_resilient: bool = False,
    ):
        self.id = id
        self.name = name
        self.pipeline_def = pipeline_def
        self.warnings = warnings
        self.is_resilient = is_resilient

    def __eq__(self, other: "Pipeline"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"Pipeline(id= {self.id}, name= {self.name!r}, pipeline_def= {self.pipeline_def!r}, warnings= {self.warnings!r}, is_resilient= {self.is_resilient})"

    def _as_payload(self):
        base = dict()
        base["id"] = str(self.id)
        base["name"] = self.name
        base["definition"] = self.pipeline_def._as_payload()
        base["warnings"] = self.warnings
        return base

    @staticmethod
    def _from_payload(payload: dict) -> Union[DtlError, "Pipeline"]:
        pipeline_id = map_option(payload.get("id"), _parse_uuid)
        if isinstance(pipeline_id, DtlError):
            return pipeline_id

        name = payload.get("name")
        if name is None:
            return DtlError("Missing parameter for Pipeline: 'name'")

        definition = payload.get("definition")
        if definition is None:
            return DtlError("Missing parameter for Pipeline: 'definition'")
        else:
            definition = PipelineDef._from_payload(definition)
            if isinstance(definition, DtlError):
                return definition

        resilient = payload.get("resilient")
        if name is None:
            resilient = False

        warnings = payload.get("warnings")
        return Pipeline(pipeline_id, name, definition, warnings, resilient)



from typing import Optional, Union, Dict
from uuid import UUID

from datalogue.errors import DtlError
from datalogue.dtl_utils import map_option, _parse_uuid


class PipelineTemplate:
    """
    A local pipeline template. Once created, this template can be
    passed mapped inputs to generate pipelines or jobs

    :param name: a name for this template
    :param template: the structure of the template, as a YAML-styled
    string. Please see https://mustache.github.io/mustache.5.html
    :param dummy_input_map: an example of data that will be used to
    validate this template upon creation, unless disabled upon the
    creation step
    :param :id: will be generated by the backend upon creation and used to
    reference this template
    """

    def __init__(
        self, name: str, template: str, dummy_input_map: Optional[Dict[str, str]] = None, id: Optional[UUID] = None,
    ):
        self.name = name
        self.template = template
        self.dummy_input_map = dummy_input_map
        self.id = id

    def __eq__(self, other):
        return (self.name == other.name) and (self.id == other.id)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(\n"
            f"id= {self.id!r},\n"
            f"name= {self.name!r},\n"
            f"template= {self.template}\n"
            f")"
        )

    def _as_payload(self) -> Union[DtlError, dict]:
        payload = {
            "name": self.name,
            "template": self.template,
            "dummy_input_map": self.dummy_input_map,
        }
        return payload

    @staticmethod
    def _from_payload(payload: dict) -> Union[DtlError, "PipelineTemplate"]:
        name = payload.get("name")
        if name is None:
            return DtlError("'name' is missing for a template")

        id = map_option(payload.get("id"), _parse_uuid)
        template = payload.get("template")
        # todo anmol - put the below fields in return instance as well
        created_by = payload.get("createdBy")
        created_at = payload.get("createdAt")
        updated_at = payload.get("updatedAt")
        last_update_by = map_option(payload.get("lastUpdateBy"), _parse_uuid)

        return PipelineTemplate(name, template, id=id)



from typing import List, Union, Optional
from datalogue.errors import DtlError, _property_not_found
from datalogue.models.transformations.commons import Transformation
from datalogue.dtl_utils import _parse_string_list


class ToInt(Transformation):
    """
      Allows to casts a node value to int
      """

    type_str = "ToInt"

    def __init__(self, path: List[str]):
        """
        Builds a Casting transformation to Int

        :param path: path to node to be cast to Int
        """
        Transformation.__init__(self, ToInt.type_str)
        self.path = path

    def __eq__(self, other: "ToInt"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"ToInt({self.path})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["path"] = self.path
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "ToInt"]:
        path = json.get("path")
        if path is None:
            return _property_not_found("path", json)

        path = _parse_string_list(path)
        if isinstance(path, DtlError):
            return path

        return ToInt(path)


class ToDouble(Transformation):
    """
      Allows to casts a node value to double
      """

    type_str = "ToDouble"

    def __init__(self, path: List[str]):
        """
        Builds a Casting transformation to Double

        :param path: path to node to be cast to Double
        """
        Transformation.__init__(self, ToDouble.type_str)
        self.path = path

    def __eq__(self, other: "ToDouble"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"ToDouble(path= {self.path})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["path"] = self.path
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "ToDouble"]:
        path = json.get("path")
        if path is None:
            return _property_not_found("path", json)

        path = _parse_string_list(path)
        if isinstance(path, DtlError):
            return path

        return ToDouble(path)


class JsonStringify(Transformation):
    """
        Selects a tree under the node of the supplied field Address.
        That tree is removed from the adg and is the replaced with a node at the same address.
        That new node has as a value the Json version of the tree that was removed.
      """

    type_str = "JsonStringify"

    def __init__(self, field_address: List[str], optional_input: bool = False):
        """
        Builds a transformation to turn a tree to json

        :param field_address: path to root of the tree to transform as a Json string
        :param optional_input: if the path is not found in the schema, will not return an error
        """
        Transformation.__init__(self, JsonStringify.type_str)
        self.field_address = field_address
        self.optional_input = optional_input

    def __eq__(self, other: "JsonStringify"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"JsonStringify(field_address= {self.field_address}, optional_input= {self.optional_input})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["fieldAddress"] = self.field_address
        base["optionalInput"] = self.optional_input
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "JsonStringify"]:
        path = json.get("fieldAddress")
        if path is None:
            return _property_not_found("fieldAddress", json)

        path = _parse_string_list(path)
        if isinstance(path, DtlError):
            return path

        optional_input = json.get("optionalInput")
        if optional_input is None:
            return _property_not_found("optionalInput", json)

        return JsonStringify(path, optional_input)


class ToDate(Transformation):
    """
    Allows to casts a node value to a date time with timezone.

    You can use the following symbols in the formatting string::

        Symbol  Meaning                     Presentation      Examples
        ------  -------                     ------------      -------
        G       era                         text              AD; Anno Domini; A
        u       year                        year              2004; 04
        y       year-of-era                 year              2004; 04
        D       day-of-year                 number            189
        M/L     month-of-year               number/text       7; 07; Jul; July; J
        d       day-of-month                number            10

        Q/q     quarter-of-year             number/text       3; 03; Q3; 3rd quarter
        Y       week-based-year             year              1996; 96
        w       week-of-week-based-year     number            27
        W       week-of-month               number            4
        E       day-of-week                 text              Tue; Tuesday; T
        e/c     localized day-of-week       number/text       2; 02; Tue; Tuesday; T
        F       week-of-month               number            3

        a       am-pm-of-day                text              PM
        h       clock-hour-of-am-pm (1-12)  number            12
        K       hour-of-am-pm (0-11)        number            0
        k       clock-hour-of-am-pm (1-24)  number            0

        H       hour-of-day (0-23)          number            0
        m       minute-of-hour              number            30
        s       second-of-minute            number            55
        S       fraction-of-second          fraction          978
        A       milli-of-day                number            1234
        n       nano-of-second              number            987654321
        N       nano-of-day                 number            1234000000

        V       time-zone ID                zone-id           America/Los_Angeles; Z; -08:30
        z       time-zone name              zone-name         Pacific Standard Time; PST
        O       localized zone-offset       offset-O          GMT+8; GMT+08:00; UTC-08:00;
        X       zone-offset 'Z' for zero    offset-X          Z; -08; -0830; -08:30; -083015; -08:30:15;
        x       zone-offset                 offset-x          +0000; -08; -0830; -08:30; -083015; -08:30:15;
        Z       zone-offset                 offset-Z          +0000; -0800; -08:00;
        julian  JDE Julian Date Format      text              120060 (equivalent to 29/02/2020, midnight time)
      """

    type_str = "ToDate"

    def __init__(
        self, path: List[str], date_format: Optional[List[str]] = None, optional_input: bool = False,
    ):
        """
        Builds a Casting transformation to a datetime with timezone

        :param path: path to node to be cast to Int
        :param date_format: List of formats to be used to attempt to parse the string. Defaults to [YYYY-MM-DDTHH:mm:ss.VVZ]
        """
        Transformation.__init__(self, ToDate.type_str)
        self.path = path
        self.date_format = date_format
        self.optional_input = optional_input

    def __eq__(self, other: "ToDate"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return (
            f"ToDate(path= {self.path!r}, date_format= {self.date_format!r}, optional_input= {self.optional_input!r})"
        )

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["path"] = self.path
        base["dateFormats"] = self.date_format
        base["optionalInput"] = self.optional_input
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "ToDate"]:
        path = json.get("path")
        dateformats = json.get("dateFormats")
        if path is None:
            return _property_not_found("path", json)

        if dateformats is None:
            return _property_not_found("dateFormats", json)

        optional_input = json.get("optionalInput")
        if optional_input is None:
            optional_input = False

        path = _parse_string_list(path)
        dateformats = _parse_string_list(dateformats)
        if isinstance(path, DtlError):
            return path
        if isinstance(dateformats, DtlError):
            return dateformats

        return ToDate(path, dateformats, optional_input)


class FormatDate(Transformation):
    """
    Formats a timestamp datapoint to a string. For help formatting your date, please follow java guidelines:
    https://docs.oracle.com/en/java/javase/11/docs/api/java.base/java/text/SimpleDateFormat.html

    :param path: path to the field to format as a list of strings, e.g. ["DOB"] targets the field "DOB". If there is
    more than one field in the dataset named as such, use a list of strings to provide an address to your target field
    :param output_date_format: output format of the date as a string following java datetime standards
    """

    type_str = "DateStringFormatter"

    def __init__(self, path: List[str], output_date_format: str):
        Transformation.__init__(self, FormatDate.type_str)
        self.path = path
        self.output_date_format = output_date_format

    def __eq__(self, other: "FormatDate"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"FormatDate(path= {self.path!r}, " f"output_date_format= {self.output_date_format!r})"

    def _as_payload(self) -> Union[dict, DtlError]:
        if not isinstance(self.path, List):
            return DtlError("Path must be a list of string describing the path to the target field")
        if not isinstance(self.output_date_format, str):
            return DtlError("Output date format must be a string, following Java guidelines")
        base = {
            "type": "DateStringFormatter",
            "path": self.path,
            "outputDateFormat": self.output_date_format,
        }
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "FormatDate"]:
        path_from_payload = json.get("path")
        output_date_format = json.get("outputDateFormat")
        if path_from_payload is None:
            return _property_not_found("path", json)

        if output_date_format is None:
            return _property_not_found("outputDateFormat", json)

        path = _parse_string_list(path_from_payload)

        if isinstance(path, DtlError):
            return path
        if isinstance(output_date_format, DtlError):
            return output_date_format

        return FormatDate(path, output_date_format,)


class ToMacAddress(Transformation):
    """
      Allows to casts a node value to MacAddress
      """

    type_str = "ToMacAddress"

    def __init__(self, path: List[str]):
        """
        Builds a Casting transformation to MacAddress
        :param path: path to node to be cast to MacAddress
        """
        Transformation.__init__(self, ToMacAddress.type_str)
        self.path = path

    def __eq__(self, other: "ToMacAddress"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"ToMacAddress({self.path})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["path"] = self.path
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "ToMacAddress"]:
        path = json.get("path")
        if path is None:
            return _property_not_found("path", json)

        path = _parse_string_list(path)
        if isinstance(path, DtlError):
            return path

        return ToMacAddress(path)

class ToUUID(Transformation):
    """
    Allows casting of a node value to a UUID
    """
    type_str = "ToUUID"

    def __init__(self, path: List[str]):
        """
        Builds a Casting transformation to UUID

        :param path: path to node to be cast to UUID
        """
        Transformation.__init__(self, ToUUID.type_str)
        self.path = path

    def __eq__(self, other: "ToUUID"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"ToUUID({self.path})"
    
    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["path"] = self.path
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "ToUUID"]:

        path = json.get("path")
        if path is None:
            return _property_not_found("path", json)

        path = _parse_string_list(path)
        if isinstance(path, DtlError):
            return path

        return ToUUID(path)



from datalogue.models.transformations.commons import Transformation, DataType
from datalogue.dtl_utils import _parse_string_list, SerializableStringEnum
from datalogue.errors import (
    _enum_parse_error,
    DtlError,
    _property_not_found,
    _invalid_property_type,
)
from typing import List, Union


class Operator(SerializableStringEnum):
    """
    Operations that can be done between 2 Operands
    """

    Add = "+"
    Sub = "-"
    Div = "/"
    Mul = "*"
    Pow = "*"
    Mod = "%"

    @staticmethod
    def parse_error(s: str) -> str:
        return _enum_parse_error("operator", s)

    @staticmethod
    def from_str(string: str) -> Union[DtlError, "Operator"]:
        return SerializableStringEnum.from_str(Operator)(string)


class Math(Transformation):
    """
    Allows to do math between nodes in the adg
    """

    type_str = "Math"

    def __init__(
        self, left: Union[List[str], float, int], op: Operator, right: Union[List[str], float, int], out: List[str],
    ):
        """
        Builds a Math transformation

        :param left: Left side of the operation. Can either be a value or a path to a value in the tree
        :param op: Operator to be used to execute the operation
        :param right: Right side of the operation. Can either be a value or a path to a value in the tree
        :param out: path where to put the result of the operation
        """
        Transformation.__init__(self, Math.type_str)
        self.left = left
        self.op = op
        self.right = right
        self.out = out

    def __eq__(self, other: "Math"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"Math(left= {self.left}, op= {self.op.value}, right= {self.right}, out= {self.out})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["operand1"] = Math._operand_as_payload(self.left)
        base["operand2"] = Math._operand_as_payload(self.right)
        base["operator"] = self.op.value
        base["resultPath"] = self.out

        return base

    def as_equation(self) -> str:
        """
        Returns the transformation written as a simple mathematical equation
        """
        return f"{self.left} {self.op.value} {self.right} = {self.out}"

    @staticmethod
    def _operand_as_payload(op: Union[List[str], float, int]) -> dict:
        if isinstance(op, list):
            return {"type": "Path", "path": op}
        else:
            return {"type": "Value", "value": float(op)}

    @staticmethod
    def _operand_from_payload(json: dict) -> Union[List[str], float, int, DtlError]:
        value = json.get("path")
        if value is not None:
            return _parse_string_list(value)

        value = json.get("value")
        if value is not None:
            return value

        return DtlError(f"Could not parse operand: {json}")

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "Math"]:
        right = json.get("operand2")
        if right is None:
            return _property_not_found("operand1", json)

        right = Math._operand_from_payload(right)
        if isinstance(right, DtlError):
            return right

        left = json.get("operand1")
        if left is None:
            return _property_not_found("operand3", json)

        left = Math._operand_from_payload(left)
        if isinstance(left, DtlError):
            return left

        op = json.get("operator")
        if op is None:
            return _property_not_found("operator", json)

        op = Operator.from_str(op)
        if isinstance(op, DtlError):
            return op

        out = json.get("resultPath")
        if out is None:
            return _property_not_found("resultPath", json)

        out = _parse_string_list(out)
        if isinstance(out, DtlError):
            return out

        return Math(left, op, right, out)



from datetime import datetime
from uuid import UUID
from typing import Union, Optional, List

from dateutil.parser import parse

from datalogue.dtl_utils import SerializableStringEnum
from datalogue.dtl_utils import _parse_list, _parse_uuid
from datalogue.errors import _enum_parse_error, DtlError


class JobStatus(SerializableStringEnum):
    Scheduled = "Scheduled"
    Defined = "Defined"
    Running = "Running"
    Succeeded = "Succeeded"
    Failed = "Failed"
    Unknown = "Unknown"
    Cancelled = "Cancelled"

    @staticmethod
    def parse_error(s: str) -> str:
        return DtlError(_enum_parse_error("job status", s))

    @staticmethod
    def job_status_from_str(string: str) -> Union[DtlError, "JobStatus"]:
        return SerializableStringEnum.from_str(JobStatus)(string)


class JobType(SerializableStringEnum):
    Pipeline = "Pipeline"
    Sampling = "Sampling"

    @staticmethod
    def parse_error(s: str) -> str:
        return DtlError(_enum_parse_error("job type", s))

    @staticmethod
    def job_type_from_str(string: str) -> Union[DtlError, "JobType"]:
        return SerializableStringEnum.from_str(JobType)(string)


class Job:
    def __init__(
        self,
        id: UUID,
        pipeline_id: Optional[UUID],
        run_at: datetime,
        status: "JobStatus",
        job_type: Optional["JobType"],
        created_by: UUID,
        errors: Optional[str],
        started_at: Optional[datetime],
        ended_at: Optional[datetime],
        percentage_progress: Optional[float],
        average_records_speed: Optional[float],
        estimated_processed_records_count: Optional[int],
        total_source_records: Optional[int],
        dropped_record_count: Optional[int],
        wait_on: List[UUID] = [],
        abort_if_overlong: bool = True,
        frequency: Optional[str] = None,
        parent_job_id: Optional[UUID] = None,
        created_by_email: Optional[str] = None,
        duration: Optional[int] = None,
        source_connection_id: Optional[UUID] = None,
        source_connection_name: Optional[str] = None,
        source_resource_name: Optional[str] = None,
        target_connection_id: Optional[UUID] = None,
        target_connection_name: Optional[str] = None,
        target_resource_name: Optional[str] = None,
        runner_id: str = "",
    ):
        self.id = id
        self.pipeline_id = pipeline_id
        self.run_at = run_at
        self.status = status
        self.job_type = job_type
        self.created_by = created_by
        self.errors = errors
        self.ended_at = ended_at
        self.started_at = started_at
        self.percentage_progress = percentage_progress
        self.average_records_speed = average_records_speed
        self.estimated_processed_records_count = estimated_processed_records_count
        self.total_source_records = total_source_records
        self.dropped_record_count = dropped_record_count
        self.wait_on = wait_on
        self.abort_if_overlong = abort_if_overlong
        self.frequency = frequency
        self.parent_job_id = parent_job_id
        self.created_by_email = created_by_email
        self.duration = duration
        self.source_connection_id = source_connection_id
        self.source_connection_name = source_connection_name
        self.source_resource_name = source_resource_name
        self.target_connection_id = target_connection_id
        self.target_connection_name = target_connection_name
        self.target_resource_name = target_resource_name
        self.runner_id = runner_id

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"id= {self.id!r}, "
            f"pipeline_id= {self.pipeline_id!r}, "
            f"status= {self.status!r}, "
            f"job_type= {self.job_type!r}, "
            f"run_at= {self.run_at!r}, "
            f"wait_on= {self.wait_on!r}, "
            f"created_by= {self.created_by!r}, "
            f"created_by_email= {self.created_by_email!r}, "
            f"errors= {self.errors!r}, "
            f"ended_at= {self.ended_at!r}, "
            f"started_at= {self.started_at!r}, "
            f"abort_if_overlong= {self.abort_if_overlong}, "
            f"frequency= {self.frequency}, "
            f"parent_job_id= {self.parent_job_id}, "
            f"percent_progress= {self.percentage_progress!r}, "
            f"average_records_speed= {self.average_records_speed!r}, "
            f"estimated_processed_records_count= {self.estimated_processed_records_count!r}, "
            f"total_source_records= {self.total_source_records!r}, "
            f"duration= {self.duration!r}, "
            f"dropped_record_count= {self.dropped_record_count!r}, "
            f"source_connection_id= {self.source_connection_id!r}, "
            f"source_connection_name= {self.source_connection_name!r}, "
            f"source_resource_name= {self.source_resource_name!r}, "
            f"target_connection_id= {self.target_connection_id!r}, "
            f"target_connection_name= {self.target_connection_name!r}, "
            f"target_resource_name= {self.target_resource_name!r},"
            f"runner_id= {self.runner_id!r})"
        )

    @classmethod
    def _from_payload(cls, json: dict) -> Union[DtlError, "Job"]:
        """
        Parses a dictionary into a Job object. Currently configured
        to parse the (latest) job data model returned by v2 jobs service.
        """
        job_id = json.get("id")
        if job_id is None:
            return DtlError("Job object should have a 'id' property")
        else:
            try:
                job_id = UUID(job_id)
            except ValueError:
                return DtlError("'id' field was not a proper uuid")

        status = json.get("status")
        if status is None:
            return DtlError("Job object should have a 'status' property")
        else:
            status = JobStatus.job_status_from_str(status)
            if isinstance(status, DtlError):
                return status

        job_type = json.get("jobType")
        if job_type is None:
            return DtlError("Job object should have a 'jobType' property")
        else:
            job_type = JobType.job_type_from_str(job_type)
            if isinstance(job_type, DtlError):
                return job_type

        run_at = json.get("scheduledAt")
        if run_at is None:
            return DtlError("Job object should have a 'runDate' property")
        else:
            try:
                run_at = parse(run_at)
            except ValueError:
                return DtlError("The 'runDate' could not be parsed as a valid date")

        created_by = json.get("owner")
        if created_by is None:
            return DtlError("Job object should have a 'owner' property")
        else:
            try:
                created_by = UUID(created_by)
            except ValueError:
                return DtlError("'owner' was not a proper uuid")

        job_type = json.get("jobType")
        if job_type is None:
            return DtlError("Job object should have a 'jobType' property")
        else:
            job_type = JobType.job_type_from_str(job_type)
            if isinstance(job_type, DtlError):
                return job_type

        # optional fields now

        ### Temporary optional fields ###
        wait_on = json.get("waitOn")
        if wait_on is not None:
            wait_on = _parse_list(_parse_uuid)(wait_on)
            if isinstance(wait_on, DtlError):
                return wait_on

        # make op metrics optional for now
        average_records_speed = json.get("averageRecordsSpeed")
        if average_records_speed is not None:
            try:
                average_records_speed = float(average_records_speed)
            except ValueError:
                return DtlError("'averageRecordsSpeed' was not a proper number")

        percentage_progress = json.get("percentage")
        if percentage_progress is not None:
            try:
                percentage_progress = float(percentage_progress)
            except ValueError:
                return DtlError("'percentage' was not a proper number")

        total_source_records = json.get("totalSourceRecords")
        if total_source_records is not None:
            try:
                total_source_records = int(total_source_records)
            except ValueError:
                return DtlError("'totalSourceRecords' was not a proper int")

        estimated_processed_records_count = json.get("records")
        if estimated_processed_records_count is not None:
            try:
                estimated_processed_records_count = int(estimated_processed_records_count)
            except ValueError:
                return DtlError("'records' was not a proper int")

        errors = json.get("details")

        parent_job_id = json.get("parentJobId")

        abort_if_overlong = json.get("abortIfOverlong")

        created_by_email = json.get("ownerEmail")

        # ended_at is optional
        ended_at = json.get("endTime")
        if ended_at is not None:
            try:
                ended_at = parse(ended_at)
            except ValueError:
                return DtlError("The 'endTime' could not be parsed as a valid date")

        # started_at is optional
        started_at = json.get("startTime")
        if started_at is not None:
            try:
                started_at = parse(started_at)
            except ValueError:
                return DtlError("The 'startTime' could not be parsed as a valid date")

        duration = json.get("duration")
        if duration is not None:
            try:
                duration = int(duration)
            except ValueError:
                return DtlError("'duration' was not a proper int")

        dropped_records = json.get("droppedRecords")
        if dropped_records is not None:
            try:
                dropped_records = int(dropped_records)
            except ValueError:
                return DtlError("'droppedRecords' was not a proper int")

        frequency = json.get("scheduleFrequency")
        if frequency is not None:
            try:
                frequency = int(frequency)
            except ValueError:
                return DtlError("'scheduleFrequency' was not a proper int")

        pipeline_id = json.get("pipelineId")
        if pipeline_id is not None:
            try:
                pipeline_id = UUID(pipeline_id)
            except ValueError:
                return DtlError("'pipelineId' field was not a proper uuid")

        source_connection_id = json.get("sourceId")
        if source_connection_id is not None:
            try:
                source_connection_id = UUID(source_connection_id)
            except ValueError:
                return DtlError("The 'sourceId' was not a proper uuid")

        source_connection_name = json.get("sourceName")
        source_resource_name = json.get("sourceResourceName")

        target_connection_id = json.get("targetId")
        if target_connection_id is not None:
            try:
                target_connection_id = UUID(target_connection_id)
            except ValueError:
                return DtlError("The 'targetId' was not a proper uuid")

        target_connection_name = json.get("targetName")
        target_resource_name = json.get("targetResourceName")
        runner_id = json.get("runnerId")

        return cls(
            id=job_id,
            pipeline_id=pipeline_id,
            run_at=run_at,
            status=status,
            job_type=job_type,
            created_by=created_by,
            errors=errors,
            started_at=started_at,
            ended_at=ended_at,
            percentage_progress=percentage_progress,
            average_records_speed=average_records_speed,
            estimated_processed_records_count=estimated_processed_records_count,
            total_source_records=total_source_records,
            duration=duration,
            dropped_record_count=dropped_records,
            wait_on=wait_on,
            abort_if_overlong=abort_if_overlong,
            frequency=frequency,
            parent_job_id=parent_job_id,
            created_by_email=created_by_email,
            source_connection_id=source_connection_id,
            source_connection_name=source_connection_name,
            source_resource_name=source_resource_name,
            target_connection_id=target_connection_id,
            target_connection_name=target_connection_name,
            target_resource_name=target_resource_name,
            runner_id=runner_id,
        )

    @classmethod
    def _from_payload_v1(cls, json: dict) -> Union[DtlError, "Job"]:
        """
        Parses a dictionary into a Job object. Currently configured
        to parse the old job data model, which is still used by the streams api.
        Should be only used within the pipeline client(for now)
        """
        job_id = json.get("jobId")
        if job_id is None:
            return DtlError("Job object should have a 'jobId' property")
        else:
            try:
                job_id = UUID(job_id)
            except ValueError:
                return DtlError("'jobId' field was not a proper uuid")

        pipeline_id = json.get("streamId")
        if pipeline_id is not None:
            try:
                pipeline_id = UUID(pipeline_id)
            except ValueError:
                return DtlError("'streamId' field was not a proper uuid")

        run_at = json.get("runDate")
        if run_at is None:
            return DtlError("Job object should have a 'runDate' property")
        else:
            try:
                run_at = parse(run_at)
            except ValueError:
                return DtlError("The 'runDate' could not be parsed as a valid date")

        status = json.get("status")
        if status is None:
            return DtlError("Job object should have a 'status' property")
        else:
            status = JobStatus.job_status_from_str(status)
            if isinstance(status, DtlError):
                return status

        remaining_time_millis = json.get("remainingTimeInMillis")
        if remaining_time_millis is not None:
            try:
                remaining_time_millis = int(remaining_time_millis)
            except ValueError:
                return DtlError("'remainingTimeInMillis' was not a proper int")

        percentage_progress = json.get("percentage")
        if percentage_progress is None:
            percentage_progress = None
        else:
            try:
                percentage_progress = float(percentage_progress)
            except ValueError:
                return DtlError("'percentage' was not a proper number")

        created_by = json.get("createdBy")
        if created_by is None:
            return DtlError("Job object should have a 'createdBy' property")
        else:
            try:
                created_by = UUID(created_by)
            except ValueError:
                return DtlError("'createdBy' was not a proper uuid")

        # details is optional
        errors = json.get("details")

        # ended_at is optional
        ended_at = json.get("endDate")
        if ended_at is not None:
            try:
                ended_at = parse(ended_at)
            except ValueError:
                return DtlError("The 'endDate' could not be parsed as a valid date")

        # started_at is optional
        started_at = json.get("startDate")
        if started_at is not None:
            try:
                started_at = parse(started_at)
            except ValueError:
                return DtlError("The 'startDate' could not be parsed as a valid date")

        wait_on = json.get("waitOn")
        if wait_on is not None:
            wait_on = _parse_list(_parse_uuid)(wait_on)
            if isinstance(wait_on, DtlError):
                return wait_on
        else:
            return DtlError("Job object should have a 'waitOn' property")

        frequency = None
        if json.get("intervalInSec") is not None:
            # TODO rename in scout so we have the same naming in both
            frequency = str(json.get("intervalInSec"))

        parent_job_id = None
        if json.get("parentJobId") is not None:
            parent_job_id = json.get("parentJobId")

        abort_if_overlong = json.get("abortIfOverlong")

        average_records_speed = json.get("averageRecordsSpeed")
        if average_records_speed is None:
            return DtlError("Job object should have a 'averageRecordsSpeed' property")
        else:
            try:
                average_records_speed = float(average_records_speed)
            except ValueError:
                return DtlError("'averageRecordsSpeed' was not a proper number")

        current_records_speed = json.get("currentRecordsSpeed")
        if current_records_speed is None:
            return DtlError("Job object should have a 'currentRecordsSpeed' property")
        else:
            try:
                current_records_speed = float(current_records_speed)
            except ValueError:
                return DtlError("'currentRecordsSpeed' was not a proper number")

        estimated_processed_records_count = json.get("estimatedProcessedRecordsCount")
        if estimated_processed_records_count is None:
            return DtlError("Job object should have a 'estimatedProcessedRecordsCount' property")
        else:
            try:
                estimated_processed_records_count = int(estimated_processed_records_count)
            except ValueError:
                return DtlError("'estimatedProcessedRecordsCount' was not a proper int")

        average_bytes_speed = json.get("averageBytesSpeed")
        if average_bytes_speed is not None:
            try:
                average_bytes_speed = float(average_bytes_speed)
            except ValueError:
                return DtlError("'averageBytesSpeed' was not a proper number")

        current_bytes_speed = json.get("currentBytesSpeed")
        if current_bytes_speed is not None:
            try:
                current_bytes_speed = float(current_bytes_speed)
            except ValueError:
                return DtlError("'currentBytesSpeed' was not a proper number")

        estimated_processed_bytes_count = json.get("estimatedProcessedBytesCount")
        if estimated_processed_bytes_count is not None:
            try:
                estimated_processed_bytes_count = int(estimated_processed_bytes_count)
            except ValueError:
                return DtlError("'estimatedProcessedBytesCount' was not a proper int")

        total_source_records = json.get("totalSourceRecords")
        if total_source_records is not None:
            try:
                total_source_records = int(total_source_records)
            except ValueError:
                return DtlError("'totalSourceRecords' was not a proper int")

        total_source_bytes = json.get("totalSourceBytes")
        if total_source_bytes is not None:
            try:
                total_source_bytes = int(total_source_bytes)
            except ValueError:
                return DtlError("'totalSourceBytes' was not a proper int")

        dropped_records = json.get("rejectedRecordsCount")
        if dropped_records is not None:
            try:
                dropped_records = int(dropped_records)
            except ValueError:
                return DtlError("'rejectedRecordsCount' was not a proper int")

        job = cls(
            id=job_id,
            pipeline_id=pipeline_id,
            job_type=None,
            run_at=run_at,
            status=status,
            percentage_progress=percentage_progress,
            created_by=created_by,
            errors=errors,
            started_at=started_at,
            ended_at=ended_at,
            average_records_speed=average_records_speed,
            estimated_processed_records_count=estimated_processed_records_count,
            total_source_records=total_source_records,
            wait_on=wait_on,
            abort_if_overlong=abort_if_overlong,
            frequency=frequency,
            parent_job_id=parent_job_id,
            dropped_record_count=dropped_records,
        )
        return job



from typing import List, Union, Optional, NamedTuple
from abc import ABC, abstractmethod
from datalogue.dtl_utils import (
    _parse_list,
    _parse_uuid,
    map_option,
    _parse_datetime,
    is_valid_uuid,
)
from uuid import UUID
from datetime import datetime
from datalogue.errors import DtlError, _property_not_found
from datalogue.models.ontology import OntologyNode


class ClassificationMethod(ABC):
    """
    Abstract class containing the three subclasses:

    - MLMethod, for classification based on a machine learning model
    - RegexOnValueMethod, for classification based on regex matching on data values
    - RegexOnFieldNameMethod, for classification based on regex matching on field names
    """

    type_field = "type"

    def __init__(self, transformation_type: str):
        self.type = transformation_type
        super().__init__()

    def _base_payload(self) -> dict:
        return dict([(ClassificationMethod.type_field, self.type)])

    @abstractmethod
    def _as_payload(self) -> dict:
        """
        Represents the Classification method as a json payload

        :return:
        """
        pass

    @staticmethod
    @abstractmethod
    def _from_payload(json: dict) -> "ClassificationMethod":
        """
        Represents the Classification method as a json payload

        :return:
        """


class RegexMap(object):
    def __init__(self, regex_ids: List[Union[UUID, str]], class_id: Union[UUID, str]):
        """
        Build a mapping between an ontological class and one or more regexes. The mapping will be used by a
            RegexOnFieldNameMethod or RegexOnValueMethod to apply classifications.

        :param class_id: the id of the ontological class that a regex match should maps to
        :param regex_ids: a list of regex ids whose matches map to the ontological class

        """
        self.regex_ids = regex_ids
        self.class_id = class_id

    def __repr__(self):
        pairs = ""
        for id in self.regex_ids:
            pairs += "\n  regex_id(" + str(id) + ") --> class_id(" + str(self.class_id) + ")"

        return "RegexMap(" f"{pairs})"

    def __eq__(self, other: "RegexMap"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def _as_payload(self) -> dict:
        return {
            "regexes": list(map(lambda i: str(i), self.regex_ids)),
            "classId": str(self.class_id),
        }

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "RegexMap"]:
        regexes = json.get("regexes")
        if regexes is None:
            return _property_not_found("regexes", json)

        class_id = json.get("classId")
        if regexes is None:
            return _property_not_found("classId", json)

        return RegexMap(regexes, class_id)


class RegexOnFieldNameMethod(ClassificationMethod):
    type_str = "RegexOnFieldNameMethod"

    def __init__(self, regex_name_maps: List[RegexMap]):
        """
        Build a classification method based on regexes applied to field names.
        This classification method will apply classes when a match is found between a data point’s field name
            and a supplied regex.

        note: the classification score supplied by a regex match is 1.0

        :param regex_name_maps: a list of the mappings between regexes and the classes they map to
        """
        ClassificationMethod.__init__(self, RegexOnFieldNameMethod.type_str)
        self.regex_name_maps = regex_name_maps

    def __repr__(self):
        pairs = repr(self.regex_name_maps)

        return (
            "\nRegexOnFieldNameMethod("
            f"{pairs}"
            f"\n  - Regexes above used to target data \033[1mfield names\033[0m for classification\n)"
        )

    def __eq__(self, other: "RegexOnFieldNameMethod"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["regexNameMaps"] = list(map(lambda m: m._as_payload(), self.regex_name_maps))
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "RegexOnFieldNameMethod"]:
        regex_name_maps = json.get("regexNameMaps")
        if regex_name_maps is None:
            return _property_not_found("regexNameMaps", json)
        else:
            regex_name_maps = _parse_list(RegexMap._from_payload)(regex_name_maps)
            if isinstance(regex_name_maps, DtlError):
                return regex_name_maps

        return RegexOnFieldNameMethod(regex_name_maps)


class RegexOnValueMethod(ClassificationMethod):
    type_str = "RegexOnValueMethod"

    def __init__(self, regex_value_maps: List[RegexMap]):
        """
        Build a classification method based on regexes applied to data values.
        This classification method will apply classes when a match is found between a data point’s value and a
         supplied regex.

        note: the classification score supplied by a regex match is 1.0

        :param regex_value_maps: a list of the mappings between regexes and the classes they map to
        """
        ClassificationMethod.__init__(self, RegexOnValueMethod.type_str)
        self.regex_value_maps = regex_value_maps

    def __repr__(self):
        pairs = repr(self.regex_value_maps)

        return (
            "\nRegexOnValueMethod("
            f"{pairs}"
            f"\n  - Regexes above used to target data \033[1mvalues\033[0m for classification\n)"
        )

    def __eq__(self, other: "RegexOnValueMethod"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["regexValueMaps"] = list(map(lambda m: m._as_payload(), self.regex_value_maps))
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "RegexOnValueMethod"]:
        regex_name_maps = json.get("regexValueMaps")
        if regex_name_maps is None:
            return _property_not_found("regexValueMaps", json)
        else:
            regex_name_maps = _parse_list(RegexMap._from_payload)(regex_name_maps)
            if isinstance(regex_name_maps, DtlError):
                return regex_name_maps

        return RegexOnValueMethod(regex_name_maps)


class ClassifiedDataPoint(NamedTuple):
    test_class: OntologyNode
    score: float
    value: Union[str, float, int, datetime, bool, None]
    field_name: Optional[str]


class ClassifierDataPoint(object):
    def __init__(self, value: Optional[str], label: Optional[str]):
        """
        Build a Classifier Data Point for testing a Classifier

        :param value: the value of the test datapoint to be classified
        :param label: can be used to add a field name to the datapoint.
            Either value or label must be supplied to build a Classifier Data Point
        """
        if value is None and label is None:
            raise ValueError("Either value or field_name must be specified for testing a Classifier Data Point")
        self.value = value
        self.label = label

    def __repr__(self):
        return (
            "Classifier("
            + (f"value={self.value!r}, " if self.value is not None else "")
            + (f"value={self.label!r}, " if self.label is not None else "")
        )

    def _as_payload(self) -> dict:
        payload = {}
        if self.value is not None:
            payload["value"] = str(self.value)
        if self.label is not None:
            payload["label"] = str(self.label)
        return payload

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "ClassifierDataPoint"]:
        value = json.get("value")
        label = json.get("label")

        if value is None and label is None:
            return DtlError("Either value or field_name must be present in the response")

        return ClassifierDataPoint(value=value, label=label)


class Classifier(object):
    def __init__(
        self,
        name: str,
        classification_methods: List[ClassificationMethod],
        description: str = "",
        default_class_id: Optional[Union[UUID, str]] = None,
        id: Optional[UUID] = None,
        owner: Optional[UUID] = None,
        editors: Optional[List[UUID]] = None,
        created_at: Optional[datetime] = None,
        created_by: Optional[UUID] = None,
        updated_at: Optional[datetime] = None,
        updated_by: Optional[UUID] = None,
        domain: Optional[List[OntologyNode]] = None,
    ):
        """
        Build a local Classifier

        :param classification_methods: ordered list of Classification Methods
        :param default_class_id: allows the user to specify the class to be defaulted to, if all classification methods
            fail. Default value is `Unknown`. (currently part of replace_class, see below)
        """
        self.name = name
        self.classification_methods = classification_methods
        self.description = description
        self.default_class_id = default_class_id
        self.id = id
        self.owner = owner
        self.editors = editors
        self.created_at = created_at
        self.created_by = created_by
        self.updated_at = updated_at
        self.updated_by = updated_by
        self.domain = domain

    def __repr__(self):
        return (
            "Classifier(" + (f"id={self.id!r}, " if self.id is not None else "") + f"name={self.name!r}, "
            f"default_class={self.default_class_id!r}, "
            f"classification_methods={self.classification_methods!r}"
            + (f", domain={self.domain!r}, " if self.domain is not None else "")
            + f")"
        )

    def __eq__(self, other: "Classifier"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def _as_payload(self) -> dict:
        payload = {
            "name": self.name,
            "description": self.description,
            "classificationMethods": list(map(lambda m: m._as_payload(), self.classification_methods)),
        }

        if self.default_class_id is not None:
            if is_valid_uuid(self.default_class_id):
                payload["defaultClassId"] = str(self.default_class_id)
            else:
                raise DtlError("default_class_id provided is not a valid UUID format.")

        if self.id is not None:
            payload["id"] = self.id

        if self.owner is not None:
            payload["owner"] = self.owner

        if self.editors is not None:
            payload["editors"] = self.editors

        if self.created_at is not None:
            payload["createdAt"] = self.created_at

        if self.updated_at is not None:
            payload["updatedAt"] = self.updated_at

        if self.updated_by is not None:
            payload["lastUpdateBy"] = self.updated_by

        return payload

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "Classifier"]:
        name = json.get("name")
        if name is None or not isinstance(name, str):
            return DtlError("'name' is missing or not a string")

        description = json.get("description")
        if description is None or not isinstance(description, str):
            return DtlError("'description' is missing or not a string")

        id = map_option(json.get("id"), _parse_uuid)
        owner = map_option(json.get("owner"), _parse_uuid)

        editors = json.get("editors")
        if editors is not None:
            editors = _parse_list(_parse_uuid)(editors)
            if isinstance(editors, DtlError):
                return DtlError(f"'editors has at least one invalid value: {json.get('editors')}", editors,)

        created_at = map_option(json.get("createdAt"), _parse_datetime)
        if isinstance(created_at, DtlError):
            return created_at

        created_by = map_option(json.get("createdBy"), _parse_uuid)
        if isinstance(created_by, DtlError):
            return created_by

        updated_at = map_option(json.get("updatedAt"), _parse_datetime)
        if isinstance(updated_at, DtlError):
            return updated_at

        updated_by = map_option(json.get("lastUpdateBy"), _parse_uuid)
        if isinstance(updated_by, DtlError):
            return updated_by

        classification_methods = json.get("classificationMethods")
        if classification_methods is None:
            return _property_not_found("classificationMethods", json)
        else:
            classification_methods = _parse_list(_classification_method_from_payload)(classification_methods)
            if isinstance(classification_methods, DtlError):
                return classification_methods

        default_class = json.get("defaultClassId")

        return Classifier(
            name=name,
            classification_methods=classification_methods,
            description=description,
            default_class_id=default_class,
            id=id,
            owner=owner,
            editors=editors,
            created_at=created_at,
            created_by=created_by,
            updated_at=updated_at,
            updated_by=updated_by,
        )


_ml_methods = dict(
    [
        (RegexOnFieldNameMethod.type_str, RegexOnFieldNameMethod._from_payload),
        (RegexOnValueMethod.type_str, RegexOnValueMethod._from_payload),
    ]
)


def _classification_method_from_payload(json: dict,) -> Union[DtlError, ClassificationMethod]:
    type_field = json.get(ClassificationMethod.type_field)
    if type_field is None:
        return DtlError("The json object doesn't have a '%s' property" % ClassificationMethod.type_field)

    parsing_function = _ml_methods.get(type_field)
    if parsing_function is None:
        return DtlError("Looks like '%s' Classification method is not handled by the SDK" % type_field)

    return parsing_function(json)
